# -*- coding: utf-8 -*-
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Exporter, ExporterChannel and ExporterCommand implementation"""

# from warnings import warn
import logging

import gevent
from gevent.queue import Queue

from mxcubecore.CommandContainer import (
    ChannelObject,
    CommandObject,
)

from .exporter import ExporterClient
from .exporter.StandardClient import PROTOCOL

__copyright__ = """ Copyright © 2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"

EXPORTER_CLIENTS = {}


def start_exporter(address, port, timeout=3, retries=1):
    """Start the exporter"""
    global EXPORTER_CLIENTS
    if (address, port) not in EXPORTER_CLIENTS:
        client = Exporter(address, port, timeout)
        EXPORTER_CLIENTS[(address, port)] = client
        client.start()
        return client
    return EXPORTER_CLIENTS[(address, port)]


class Exporter(ExporterClient.ExporterClient, object):
    """Exporter class"""

    STATE_EVENT = "State"
    STATUS_EVENT = "Status"
    VALUE_EVENT = "Value"
    POSITION_EVENT = "Position"
    MOTOR_STATES_EVENT = "MotorStates"

    STATE_READY = "Ready"
    STATE_INITIALIZING = "Initializing"
    STATE_STARTING = "Starting"
    STATE_RUNNING = "Running"
    STATE_MOVING = "Moving"
    STATE_CLOSING = "Closing"
    STATE_REMOTE = "Remote"
    STATE_STOPPED = "Stopped"
    STATE_COMMUNICATION_ERROR = "Communication Error"
    STATE_INVALID = "Invalid"
    STATE_OFFLINE = "Offline"
    STATE_ALARM = "Alarm"
    STATE_FAULT = "Fault"
    STATE_UNKNOWN = "Unknown"

    def __init__(self, address, port, timeout=3, retries=1):
        super(Exporter, self).__init__(address, port, PROTOCOL.STREAM, timeout, retries)

        self.started = False
        self.callbacks = {}
        self.events_queue = Queue()
        self.events_processing_task = None

    def start(self):
        """Start"""

    def stop(self):
        """Stop"""
        self.disconnect()

    def execute(self, *args, **kwargs):
        """Execute"""
        ret = ExporterClient.ExporterClient.execute(self, *args, **kwargs)
        return self._to_python_value(ret)

    def get_state(self):
        """Read the state"""
        return self.execute("getState")

    def read_property(self, *args, **kwargs):
        """Read a property"""
        ret = ExporterClient.ExporterClient.read_property(self, *args, **kwargs)
        return self._to_python_value(ret)

    def reconnect(self):
        """Reconnect"""
        return

    def on_disconnected(self):
        """Actions on disconnect"""

    def register(self, name, cb):
        """Register to a callback"""
        if callable(cb):
            self.callbacks.setdefault(name, []).append(cb)
        if not self.events_processing_task:
            self.events_processing_task = gevent.spawn(self.process_events_from_queue)

    def _to_python_value(self, value):
        """Convert exporter value to python one
        Args:
            value (str): String from the exporter
        """
        if value is None:
            return value

        if "\x1f" in value:
            value = self.parse_array(value)
            try:
                value = list(map(int, value))
            except (TypeError, ValueError):
                try:
                    value = list(map(float, value))
                except (TypeError, ValueError):
                    pass
        else:
            if value == "false":
                value = False
            elif value == "true":
                value = True
            else:
                try:
                    value = int(value)
                except (TypeError, ValueError):
                    try:
                        value = float(value)
                    except (TypeError, ValueError):
                        pass
        return value

    def on_event(self, name, value, timestamp):
        """Put the event in the queue
        Args:
            name: Name
            value: Value
            timestamp: Timestamp
        """
        self.events_queue.put((name, value))

    def process_events_from_queue(self):
        """Process events from the queue"""
        while True:
            try:
                name, value = self.events_queue.get()
            except Exception:
                return

            for cb in self.callbacks.get(name, []):
                try:
                    cb(self._to_python_value(value))
                except Exception:
                    msg = "Exception while executing callback {} for event {}".format(
                        cb, name
                    )
                    logging.exception(msg)
                    continue


class ExporterCommand(CommandObject):
    """Command implementation for Exporter"""

    def __init__(
        self, name, command, username=None, address=None, port=None, timeout=3, **kwargs
    ):
        CommandObject.__init__(self, name, username, **kwargs)
        self.command = command
        self.__exporter = start_exporter(address, port, timeout)
        msg = "Attaching Exporter command: {} {}".format(address, name)
        logging.getLogger("HWR").debug(msg)

    def __call__(self, *args, **kwargs):
        self.emit("commandBeginWaitReply", (str(self.name()),))

        try:
            ret = self.__exporter.execute(self.command, args, kwargs.get("timeout", -1))
        except Exception:
            self.emit("commandFailed", (-1, self.name()))
            raise
        else:
            self.emit("commandReplyArrived", (ret, str(self.name())))
            return ret

    def abort(self):
        """Abort"""

    def get_state(self):
        """FRead the state
        Returns:
            (str): State
        """
        return self.__exporter.get_state()

    def is_connected(self):
        """Check if connected.
        Returns:
            (bool): True if connected
        """
        return self.__exporter.is_connected()


class ExporterChannel(ChannelObject):
    """Channel implementation for Exporter"""

    def __init__(
        self,
        name,
        attribute_name,
        username=None,
        address=None,
        port=None,
        timeout=3,
        **kwargs,
    ):
        ChannelObject.__init__(self, name, username, **kwargs)

        self.__exporter = start_exporter(address, port, timeout)
        self.attribute_name = attribute_name
        self.value = None

        self.__exporter.register(attribute_name, self.update)

        msg = "Attaching Exporter channel: {} {} ".format(address, name)
        logging.getLogger("HWR").debug(msg)

        self.update()

    def update(self, value=None):
        """Emit signal update when value changed"""
        value = value or self.get_value()
        if isinstance(value, tuple):
            value = list(value)

        self.value = value
        self.emit("update", value)

    def get_value(self):
        """Get the value
        Returns:
            (str): The value
        """
        value = self.__exporter.read_property(self.attribute_name)
        return value

    def set_value(self, value):
        """Set a value
        Args:
            value (str): Value to set
        """
        self.__exporter.write_property(self.attribute_name, value)

    def is_connected(self):
        """Check if connected.
        Returns:
            (bool): True if connected.
        """
        return self.__exporter.is_connected()



from typing import Optional

from tango import (
    DevFailed,
    DeviceProxy,
)

from mxcubecore.BaseHardwareObjects import HardwareObject

"""
The sample changer maintenance hardware object for ISARA2 robot.

Provide commands for powering on and off the robot. Populates the 'equipment' tab
with a handful of other useful command, while operating the sample changer.

This hardware object assumes that ISARA2 sample changer is exposed
via tango device server implementation as implemented here:
https://gitlab.com/MaxIV/isara/dev-maxiv-isara-ns-09

Example XML configuration for this hardware object:

  <object class="ISARAMaint">
     <tangoname>b312-e/ctl/sm-01</tangoname>
     <polling>400</polling>
  </object>

  tangoname - the sample changers tango device to use
  polling   - optional, sets the polling frequency of device attributes, in milliseconds
"""

DEFAULT_POLLING = 1000


def _reword_isara_error(err_message: str) -> str:
    """
    Reword some of the ISARA error messages to make them easier to understand for the user.
    """
    msg = err_message.lower()
    if msg == "remote mode requested":
        return "it is not in remote mode"

    if msg == "doors must be closed":
        return "hutch is not searched"

    # let's use error as-is
    return msg


def _get_isara_command_error(ex: DevFailed) -> Optional[str]:
    """
    Check if specified exception encodes an ISARA command error.

    Returns the human-readable error message from ISARA, or None
    if this is not an ISARA command error exception.
    """

    # The ISARA command errors are signaled by raising DevError exception,
    # with reason set to 'ISARACommandError' and 'desc' to the error message
    # from ISARA. The DevError will be wrapped inside DevFailed exception by
    # tango.
    #
    # Check if provided exception contains an DevError signalling ISARA command error.
    for err in ex.args:
        if err.reason == "ISARACommandError":
            return _reword_isara_error(err.desc)

    # this is not an ISARA command error exception
    return None


class ISARAMaint(HardwareObject):
    def __init__(self, name):
        super().__init__(name)

        self._commands_state = dict(
            PowerOn=False,
            PowerOff=False,
            openLid=True,
            closeLid=True,
            home=True,
            dry=True,
            soak=True,
            clearMemory=True,
            reset=True,
            back=True,
            abort=True,
        )

        self._powered = None
        self._position_name = None
        self._message = None

    def init(self):
        self.isara_dev = DeviceProxy(self.tangoname)

        polling = self._get_polling()

        self._poll_attribute("Powered", polling, self._powered_updated)
        self._poll_attribute("PositionName", polling, self._position_name_updated)
        self._poll_attribute("Message", polling, self._message_updated)

    def _get_polling(self):
        """
        get polling frequency to use for device attribute poller
        """
        try:
            # polling is specified in the XML file
            return self.polling
        except AttributeError:
            # no polling is specified in the XML, use default polling value
            return DEFAULT_POLLING

    def _poll_attribute(self, attr_name: str, polling: int, callback):
        channel = self.add_channel(
            {
                "type": "tango",
                "name": f"_chn{attr_name}",
                "tangoname": self.tangoname,
                "polling": polling,
            },
            attr_name,
        )

        channel.connect_signal("update", callback)

    def _update_state(self):
        for attr in [self._powered, self._position_name, self._message]:
            if attr is None:
                # some of the values are still unknown,
                # wait until we get all values
                return

        #
        # update 'power' commands
        #
        self._commands_state["PowerOn"] = not self._powered
        self._commands_state["PowerOff"] = self._powered

        #
        # update 'positions' commands
        #
        if not self._powered or self._position_name == "undefined":
            # when powered off or running a trajectory,
            # disable position commands
            self._commands_state["home"] = False
            self._commands_state["dry"] = False
            self._commands_state["soak"] = False
        else:
            self._commands_state["home"] = True
            self._commands_state["dry"] = True
            self._commands_state["soak"] = True

            if self._position_name in ["home", "dry", "soak"]:
                # can't move to same position
                self._commands_state[self._position_name] = False

        self._emit_global_state_changed()

    def _powered_updated(self, powered):
        self._powered = powered
        self._update_state()

    def _position_name_updated(self, position_name):
        self._position_name = position_name.lower()
        self._update_state()

    def _message_updated(self, message):
        self._message = message
        self._update_state()

    def _emit_global_state_changed(self):
        self.emit(
            "globalStateChanged",
            (self._commands_state, self._message),
        )

    def _toggle_power(self, power_on: bool):
        """
        issue powerOn or powerOff commands, catching ISARA command error exception
        """
        command = self.isara_dev.PowerOn if power_on else self.isara_dev.PowerOff

        try:
            command()
        except DevFailed as ex:
            isara_err = _get_isara_command_error(ex)

            # this is not an ISARA command error, pass on the exception
            if isara_err is None:
                raise ex

            state = "on" if power_on else "off"
            raise RuntimeError(f"Can't power {state} sample changer, {isara_err}.")

    def send_command(self, cmd_name, _args=None):
        if cmd_name == "PowerOn":
            self._toggle_power(True)
        elif cmd_name == "PowerOff":
            self._toggle_power(False)
        elif cmd_name == "openLid":
            self.isara_dev.OpenLid()
        elif cmd_name == "closeLid":
            self.isara_dev.CloseLid()
        elif cmd_name == "home":
            self.isara_dev.Home()
        elif cmd_name == "dry":
            self.isara_dev.Dry()
        elif cmd_name == "soak":
            self.isara_dev.Soak()
        elif cmd_name == "clearMemory":
            self.isara_dev.ClearMemory()
        elif cmd_name == "reset":
            self.isara_dev.Reset()
        elif cmd_name == "back":
            self.isara_dev.Back()
        elif cmd_name == "abort":
            self.isara_dev.Abort()
        else:
            raise Exception(f"ISARA MAINT: unexpected command '{cmd_name}'")

    def get_global_state(self):
        return dict(glob_state="dummy"), self._commands_state, self._message

    def get_cmd_info(self):
        return [
            [
                "Power",
                [
                    ["PowerOn", "PowerOn", "Switch Power On"],
                    ["PowerOff", "PowerOff", "Switch Power Off"],
                ],
            ],
            [
                "Lid",
                [
                    ["openLid", "Open Lid", "Open Lid"],
                    ["closeLid", "Close Lid", "Close Lid"],
                ],
            ],
            [
                "Positions",
                [
                    ["home", "Home", "Actions", "Home (trajectory)"],
                    ["dry", "Dry", "Actions", "Dry (trajectory)"],
                    ["soak", "Soak", "Actions", "Soak (trajectory)"],
                ],
            ],
            [
                "Recovery",
                [
                    [
                        "clearMemory",
                        "Clear Memory",
                        "Clear Info in Robot Memory "
                        " (includes info about sample on Diffr)",
                    ],
                    [
                        "reset",
                        "Reset Fault",
                        "Acknowledge security fault",
                    ],
                    [
                        "back",
                        "Back",
                        "Return sample back to Dewar",
                    ],
                ],
            ],
            ["Abort", [["abort", "Abort", "Abort running trajectory"]]],
        ]



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
import time

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractDetector import AbstractDetector

__author__ = "Vicente Rey"
__credits__ = ["SOLEIL"]
__version__ = "2.3."
__category__ = "General"


class PX1Pilatus(AbstractDetector, HardwareObject):
    """Detector class. Contains all information about detector
       - states are 'OK', and 'BAD'
       - status is busy, exposing, ready, etc.
       - physical property is RH for pilatus, P for rayonix
    """

    def __init__(self, name):
        AbstractDetector.__init__(self)
        HardwareObject.__init__(self, name)

        self.distance_motor_hwobj = None
        self.default_distance = None
        self.default_distance_limits = None

        self.exp_time_limits = None

        self.headers = {}

    def init(self):
        self.distance_motor_hwobj = self.get_object_by_role("detector_distance")
        self.devname = self.get_property("tangoname")

        self.state_chan = self.get_channel_object("state")
        self.state_chan.connect_signal("update", self.state_changed)

        self.threshold_chan = self.get_channel_object("threshold")
        self.threshold_chan.connect_signal("update", self.threshold_changed)

        self.set_energy_cmd = self.get_command_object("set_energy")
        self.set_header_cmd = self.get_command_object("set_header")

        exp_time_limits = self.get_property("exposure_limits")
        self.exp_time_limits = map(float, exp_time_limits.strip().split(","))

    def state_changed(self, state):
        self.current_state = state

    def threshold_changed(self, threshold):
        self.current_threshold = threshold

    def prepare_acquisition(self):
        pass

    def start_acquisition(self):
        # _dir_path = self.dcpars['fileinfo']['directory'].replace('RAW_DATA','PROCESSED_DATA')
        # chmod_dir(_dir_path)
        # write_goimg(_dir_path)
        pass

    def stop_acquisition(self):
        pass

    def get_state(self):
        return self.state_chan.get_value()

    def read_state(self):
        return str(self.get_state())

    def is_fault_state(self):
        return str(self.get_state()) == "FAULT"

    def get_threshold(self):
        return self.threshold_chan.get_value()

    def get_threshold_gain(self):
        return None

    def has_shutterless(self):
        """Return True if has shutterless mode"""
        return True

    def get_beam_position(self, distance=None, wavelength=None):
        """Returns beam center coordinates"""
        beam_x = 0
        beam_y = 0
        try:
            if self.chan_beam_xy is not None:
                value = self.chan_beam_xy.get_value()
                beam_x = value[0]
                beam_y = value[1]
        except Exception:
            pass
        return beam_x, beam_y

    def get_manufacturer(self):
        return "Dectris"

    def get_model(self):
        return self.get_property("model")

    def get_detector_type(self):
        return self.get_property("type")

    def get_default_exposure_time(self):
        return self.get_property("default_exposure_time")

    def get_minimum_exposure_time(self):
        return self.get_property("minimum_exposure_time")

    def get_exposure_time_limits(self):
        """Returns exposure time limits as list with two floats"""
        return self.exp_time_limits

    def get_file_suffix(self):
        return self.get_property("file_suffix")

    def get_pixel_size(self):
        return self.get_property("px"), self.get_property("py")

    # methods for data collection
    def prepare_collection(self, dcpars):
        energy = dcpars["energy"]
        self.do_energy_calibration(energy)

        # osc_seq = dcpars['oscillation_sequence'][0]
        # file_pars = dcpars['fileinfo']

        # basedir = file_pars['directory']
        # prefix  =  "%s_%s_" % (file_pars['prefix'], file_pars['run_number'])
        #
        # first_img_no = osc_seq['start_image_number']
        # nb_frames =  osc_seq['number_of_images']
        # exp_time = osc_seq['exposure_time']
        #
        # fileformat =  "CBF"
        # trig_mode = "EXTERNAL_TRIGGER"
        # latency_time = 0.023
        #
        # logging.getLogger("HWR").debug(" Preparing detector (dev=%s) for data collection" % self.devname)
        #
        # logging.getLogger("HWR").debug("    /saving directory: %s" % basedir)
        # logging.getLogger("HWR").debug("    /prefix          : %s" % prefix)
        # logging.getLogger("HWR").debug("    /saving_format   : %s" % fileformat)
        # logging.getLogger("HWR").debug("    /trigger_mode    : %s" % trig_mode)
        # logging.getLogger("HWR").debug("    /acq_nb_frames   : %s" % nb_frames)
        # logging.getLogger("HWR").debug("    /acq_expo_time   : %s" % exp_time)
        # logging.getLogger("HWR").debug("    /latency_time    : %s" % latency_time)
        #
        # self.device.write_attribute('saving_mode', 'AUTO_FRAME')
        # self.device.write_attribute('saving_directory', basedir)
        # self.device.write_attribute('saving_prefix', prefix)
        # self.device.write_attribute('saving_format', fileformat)
        # self.device.write_attribute('saving_next_number', first_img_no)

        return True

    def start_collection(self):
        self.wait_energy_calibration()
        self.start_acquisition()

    def stop_collection(self):
        self.stop_acquisition()

    def set_image_headers(self, image_headers):
        for _header in image_headers:
            try:
                _str_header = _header[0] % _header[1]
            except Exception:
                _str_header = _header[0]

            self.set_header_cmd(_str_header)

    def do_energy_calibration(self, energy):
        energy = float(energy)
        logging.info("<PX1Pilatus> do_energy_calibration for %.4f KeV" % energy)

        PILATUS_THRESHOLD_MIN = 3774.0  # en eV
        ENERGY_CALIBRATION_MIN = 7.6  # en keV

        current_threshold = self.threshold_chan.get_value()

        energy_diff = energy - 2 * current_threshold / 1000.0

        if (
            current_threshold == PILATUS_THRESHOLD_MIN
            and energy < ENERGY_CALIBRATION_MIN
        ):
            logging.warning(
                "PX1Pilatus. Re-calibration of Pilatus detector not possible: THRESHOLD_MIN condition."
            )
            return False

        if energy_diff < (-0.08 * (2 * current_threshold / 1000.0)) or energy_diff > (
            0.05 * (2 * current_threshold / 1000.0)
        ):

            if self.read_state() != "STANDBY":
                logging.getLogger("user_level_log").error(
                    "Re-calibration of Pilatus detector not possible."
                )
                return False

            logging.info(
                "<PX1Pilatus> sending energy value of %5d KeV" % int(energy * 1000)
            )
            self.set_energy_cmd(int(energy * 1000))

            time.sleep(1)

            if self.read_state() != "STANDBY":
                logging.getLogger("user_level_log").info(
                    "Calibration of Pilatus detector in progress (takes about 1 minute)."
                )

    def wait_energy_calibration(self):
        _state = self.read_state()

        t0 = time.time()
        last_msg_time = 0
        while _state != "STANDBY":
            time.sleep(2)
            elapsed = time.time() - t0
            _state = self.read_state()
            if elapsed - last_msg_time > 10.0:
                logging.getLogger("user_level_log").info(
                    "   -  calibration in progress (elapsed time %s)." % elapsed
                )
                last_msg_time = elapsed
            logging.info(
                "    -  <PX1Pilatus>  waiting for energy calibration: %s" % _state
            )


def test_hwo(hwo):
    print(("Detector Distance is: %s " % hwo.distance.get_value()))
    print(("         state is: %s, %s" % (hwo.get_state(), type(hwo.get_state()))))
    print(("      is in fault: %s" % hwo.is_fault_state()))
    # print "going to 490 : ", hwo.distance.set_value(490)
